/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Simple Timer + Clocks.
 *
 * The Initial Developer of the Original Code is
 * George Bradt.
 *
 * Portions created by the Initial Developer are Copyright (C) 2008
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

var SimpleClocksLocations = {
    stringBundle: null,
    countriesTimeZonesDoc: null,
    countries: [],
    countryCodes: [],
    zoneName: [],
    zoneOffset: [],
    zoneFormat: [],
    zoneComment: [],
    zoneRule: [],
    selectedCountryIndex: 0,
    saveStart: "",
    saveEnd: "",
    letterStart: "",
    letterEnd: "",
    deleteNotsForLocation: [],
    notifyTable: [],
    notification: {completed: false,
                   recurring: false,
                   timeDisplay: "",
                   description: "",
                   time24hr: "",
                   timeMilli: 0,
                   day: -1,
                   url: "",
                   location: "",
                   sortValue: 0
                  },
    selectedDisplay: {location: "",
                      offset: "",
                      dstStart: "",
                      dstEnd: "",
                      localTimeZone: ""
                     },
    selectedLocationsTable: [],
    selectedLocation: {countryCode: "",
                       timezone: "",
                       location: "",
                       rule: "",
                       observesDst: false,
                       currentlyDst: false,
                       offsetMilli: 0,
                       dstOffsetMilli: 0,
                       localTimeZone: "",
                       dstLocalTimeZone: "",
                       dstStartMilli: [0, 0, 0],
                       dstEndMilli: [0, 0, 0]
                      },
    separator: "\x1D",   // used to separate selectedLocation JSON
                         // strings in preference.

    // Displays the time zone dialog.

    displayDlg: function() {
        window.open(
                "chrome://simpletimerClocks/content/simpletimerClocks-timezones.xul",
                "",
                "centerscreen, chrome, resizable");
    },

    // Called when time zone dialog is loaded.

    onLoad: function() {
        var strbundle = document.getElementById("simclk-strings");
        var prefs = Components.classes["@mozilla.org/preferences-service;1"]
                .getService(Components.interfaces.nsIPrefService)
                .getBranch("extensions.simpletimerClocks@grbradt.org.");

        var Application = Components.classes["@mozilla.org/fuel/application;1"].
                getService(Components.interfaces.fuelIApplication);

        // Make sure the pref has been initialized.
        if ( !prefs.prefHasUserValue("tzVersion") ) {
            alert(strbundle.getString("alert.error.reading.tzversion.pref"));
            return;
        }

        // Display current tz data version.
        document.getElementById("simclk-descTzVersion").value =
        		strbundle.getString("descr.tzVersion") + " " +
                prefs.getCharPref("tzVersion");

        // If a location is deleted that has nots, the nots are marked completed.
        if ( Application.storage.has("notifyTable") ) {
            this.notifyTable = Application.storage.get("notifyTable", "ERR");

            if ( this.notifyTable === "ERR" ) {
                alert(strbundle.getString("alert.error.loading.notifications"));
                return;
            }
        }

        this.countriesTimeZonesDoc = Application.storage.get("xmlCountriesTimeZonesDoc", "No doc");

        if ( this.countriesTimeZonesDoc === "No doc" ) {
            this.readCountriesTimeZonesFile("load");
        }
        else {
            this.parseXMLCountryNames();
        }

        // Display locations user already tracking.
        this.populateSelectedTree("onLoad");
    },

    // Get tz data version file.
    // Write to tzVersion pref, if changed the pref observer routine
    // initiates location update processing. Called from SimpleClocks.

    readTzVersionFile: function(mode) {
        var strbundle = document.getElementById("simclk-strings");
        var prefs = Components.classes["@mozilla.org/preferences-service;1"]
                .getService(Components.interfaces.nsIPrefService)
                .getBranch("extensions.simpletimerClocks@grbradt.org.");

        var req = new XMLHttpRequest();

//        var url = "file:///C:/Documents and Settings/Geo/My Documents/My Firefox Extensions/tz database/XML/tzVersion.xml";
        var url = "http://downloads.mozdev.org/simpletimerclocks/tzVersion.xml";

        // "true" == async.
        req.open("GET", url, true);

        // Use an anonymous function to process the request.
        req.onreadystatechange = function () {
            try {
                if ( req.readyState === 4 ) {
                    // The full server response is received.
                    // For local/chrome files, check status = 0.
                    // For remote files, check status = 200.
                    if ( req.status === 200 ) {
//                    if ( req.status === 0 ) {
                        // No problem with the request.
                        // Return the response as an XMLDocument object.
                        var tzVersionDoc = req.responseXML;
                        var exp = "/version";
                        var version = SimpleClocksLocations.getXMLSingleNodeValue(exp, tzVersionDoc);

                        if ( version > prefs.getCharPref("tzVersion") ) {
                            prefs.setCharPref("tzVersion", version);
                        }
                    }
                    else {
                        alert(strbundle.getString("alert.error.loading.tzversion") + " " + req.status);
                    }
                }
            }
            catch(e) {
                alert(strbundle.getString("alert.error.caught.exception2") + "\n" +
                      strbundle.getString("alert.error.caught.exception.msg") + " " + e.message);
            }
        };

        // Not POSTing, so parm is null.
        try {
            req.send(null);
        }
        catch(e) {
            alert(strbundle.getString("alert.error.caught.exception2") + "\n" +
                  strbundle.getString("alert.error.caught.exception.msg") + " " + e.message);
        }
    },

    // Get the countries/timezones file.
    // mode="update" when tz data files have been updated,
    // and we are updating users locations pref.

    readCountriesTimeZonesFile: function(mode) {
        var strbundle = document.getElementById("simclk-strings");
        var prefs = Components.classes["@mozilla.org/preferences-service;1"]
                .getService(Components.interfaces.nsIPrefService)
                .getBranch("extensions.simpletimerClocks@grbradt.org.");

        var Application = Components.classes["@mozilla.org/fuel/application;1"].
                getService(Components.interfaces.fuelIApplication);

        var req = new XMLHttpRequest();
//        var url = "file:///C:/Documents and Settings/Geo/My Documents/My Firefox Extensions/tz database/XML/simpletimerClocks" +
        var url = "http://downloads.mozdev.org/simpletimerclocks/simpletimerClocks" +
                prefs.getCharPref("tzVersion") + ".xml";

        req.open("GET", url, true);

        req.onreadystatechange = function () {
            try {
                if ( req.readyState === 4 ) {
//                    if ( req.status === 0 ) {
                    if ( req.status === 200 ) {
                        // Return the response as an XMLDocument object.
                        SimpleClocksLocations.countriesTimeZonesDoc = req.responseXML;
                        Application.storage.set("xmlCountriesTimeZonesDoc", SimpleClocksLocations.countriesTimeZonesDoc);

                        if ( mode == "update" ) {
                            // Trigger pref observer. Should have just used a timestamp!
                            prefs.setBoolPref("xmlRetrieved", !prefs.getBoolPref("xmlRetrieved"));
                            return;
                        }

                        try {
                            SimpleClocksLocations.parseXMLCountryNames();
                        }
                        catch(e) {
                            alert(strbundle.getString("alert.error.caught.exception1") + "\n" +
                                  strbundle.getString("alert.error.caught.exception.msg") + " " + e.message);
                        }
                    }
                    else {
                        alert(strbundle.getString("alert.error.loading.timezones") + " " + req.status);
                    }
                }
            }
            catch(e) {
                alert(strbundle.getString("alert.error.caught.exception2") + "\n" +
                      strbundle.getString("alert.error.caught.exception.msg") + " " + e.message);
            }
        };

        try {
            req.send(null);
        }
        catch(e) {
            alert(strbundle.getString("alert.error.caught.exception2") + "\n" +
                  strbundle.getString("alert.error.caught.exception.msg") + " " + e.message);
        }
    },

    // Parse XML file for country info.

    parseXMLCountryNames: function () {
        // An array of all countries.
        var exp = "/countrytimezones/country/@name";
        this.countries = this.getXMLNodeValues(exp, this.countriesTimeZonesDoc);

        // An array of all country codes.
        exp = "/countrytimezones/country/@code";
        this.countryCodes = this.getXMLNodeValues(exp, this.countriesTimeZonesDoc);

        // Populate the country listbox.
        this.populateCountryList(this.countries, this.countryCodes);
    },

    // XPath query returning nodeset values in array.

    getXMLNodeValues: function(exp, doc) {
        var resultArray = [];
        var nodeSet = doc.evaluate(exp, doc, null, XPathResult.ANY_TYPE, null);
        var node = nodeSet.iterateNext();

        while ( node )
        {
            if ( node.nodeValue ) {
                resultArray.push(node.nodeValue);
            }
            else {
                resultArray.push("N/A");
            }

            node = nodeSet.iterateNext();
        }

        return resultArray;
     },

     // XPath query returning single node value.

    getXMLSingleNodeValue: function(exp, doc) {
        var node = doc.evaluate(exp, doc, null, XPathResult.FIRST_ORDERED_NODE_TYPE, null);
        if ( node ) {
            return node.singleNodeValue.textContent;
        }
        else {
            return null;
        }
     },

     // Populate Country listbox.

     populateCountryList: function(countries, codes) {
        var lboxCountries = document.getElementById("simclk-lboxCountries");

        // Save country code for pref.
        for ( var i in countries ) {
            lboxCountries.appendItem(countries[i]);
            lboxCountries.getItemAtIndex(i).value = codes[i];
        }
    },

    // Populate Selected tree from locations table.

    populateSelectedTree: function(mode) {
        // locationsTable is built at startup and when user clicks OK.
        var strbundle = document.getElementById("simclk-strings");
        var Application = Components.classes["@mozilla.org/fuel/application;1"].
                getService(Components.interfaces.fuelIApplication);

        // An array of location objects.
        var locationsTable = Application.storage.get("locationsTable", "ERR");

        if ( locationsTable === "ERR" ) {
            alert(strbundle.getString("alert.error.loading.locations"));
            return;
        }

        var treeChildrenSelected = document.getElementById("simclk-treeSelectedItems");

        // Clear the old tree data.
        this.onDeleteAll("simclk-treeSelectedItems", mode);

        if ( mode == "reset") {
            // If any locations were deleted prior to Reset, don't delete
            // their associated nots.
            this.deleteNotsForLocation = [];
        }

        for ( var i in locationsTable ) {
            // Tree row represented by selectedDisplay object.
            this.selectedLocation = locationsTable[i];
            this.selectedDisplay.location = this.selectedLocation.location;

            if ( this.selectedLocation.observesDst ) {
                this.buildDstDisplay();
                this.selectedDisplay.offset = this.currentlyDst(this.selectedLocation, 0) ?
                         this.convertMilliToClock(this.selectedLocation.dstOffsetMilli) + "**" :
                         this.convertMilliToClock(this.selectedLocation.offsetMilli);
                this.selectedDisplay.localTimeZone = this.currentlyDst(this.selectedLocation, 0) ?
                        this.selectedLocation.dstLocalTimeZone : this.selectedLocation.localTimeZone;
            }
            else {
                this.selectedDisplay.offset = this.convertMilliToClock(this.selectedLocation.offsetMilli);
                this.selectedDisplay.dstStart = strbundle.getString("msg.not.applicable");
                this.selectedDisplay.dstEnd = strbundle.getString("msg.not.applicable");
                this.selectedDisplay.localTimeZone = this.selectedLocation.localTimeZone;
            }

            this.pushCellData(treeChildrenSelected);
        }
    },

    // Called when user selects an entry in the Countries listbox.
    // Populate the Time Zones tree.

    onCountrySelected: function() {
        var treeChildren = document.getElementById("simclk-treeTimeZonesItems");
        var listBox = document.getElementById("simclk-lboxCountries");
        var addData = [];

        // Reset zone info arrays.
        this.zoneName.length = 0;
        this.zoneOffset.length = 0;
        this.zoneFormat.length = 0;
        this.zoneComment.length = 0;
        this.zoneRule.length = 0;

        // Save the index and value (country code) of selected country.
        this.selectedCountryIndex = listBox.currentIndex;
        this.selectedLocation.countryCode = listBox.getSelectedItem(0).value;

        // Get the zone info for selected country.
        // Xpath doesn't use zero-based indexing.
        var index = this.selectedCountryIndex + 1;

        var exp = "/countrytimezones/country[" + index + "]/timezone/@name";
        this.zoneName = this.getXMLNodeValues(exp, this.countriesTimeZonesDoc);

        exp = "/countrytimezones/country[" + index + "]/timezone/@offset";
        this.zoneOffset = this.getXMLNodeValues(exp, this.countriesTimeZonesDoc);

        exp = "/countrytimezones/country[" + index + "]/timezone/@format";
        this.zoneFormat = this.getXMLNodeValues(exp, this.countriesTimeZonesDoc);

        exp = "/countrytimezones/country[" + index + "]/timezone/@comment";
        this.zoneComment = this.getXMLNodeValues(exp, this.countriesTimeZonesDoc);

        exp = "/countrytimezones/country[" + index + "]/timezone/@rule";
        this.zoneRule = this.getXMLNodeValues(exp, this.countriesTimeZonesDoc);

        // Clear old tree data.
        this.onDeleteAll("simclk-treeTimeZonesItems", "");

        // Populate time zone tree.
        for ( var i in this.zoneName ) {
            addData.push(this.zoneName[i]);
            addData.push(this.zoneOffset[i]);
            addData.push(this.zoneComment[i]);
            this.addTreeRow(treeChildren, addData);
            addData.length = 0;
        }
    },

    // Called to move selection from time zone tree to selected tree.

    onAdd: function(event) {
        var strbundle = document.getElementById("simclk-strings");
        var treeZones = document.getElementById("simclk-treeTimeZones");
        var index = treeZones.currentIndex;
        var treeSelected = document.getElementById("simclk-treeSelected");
        var treeChildrenSelected = document.getElementById("simclk-treeSelectedItems");
        var treeItems;

        if ( event.detail == 2 ) {
            // Double-click from Timezones table.
            // Get the row, col and child element at the point of mouse click.
            var tbo = treeZones.treeBoxObject;
            var row = { }, col = { }, child = { };
            tbo.getCellAt(event.clientX, event.clientY, row, col, child);

            if ( !col.value ) {
                return;
            }
        }
        else if ( index < 0 ) {
            // User didn't make a selection.
            alert(strbundle.getString("alert.warning.no.timezone.selected"));
            return;
        }

        // Save zone info. Last segment of tz timezone is city.
        this.selectedLocation.timezone = this.zoneName[index];
        var city = this.zoneName[index].substr(this.zoneName[index].lastIndexOf("/") + 1);
        this.selectedDisplay.location = city.replace(/_/g, " ");
        this.selectedLocation.location = this.selectedDisplay.location;

        // Check if the new entry is a duplicate.
        if ( treeChildrenSelected.hasChildNodes() ) {
            treeItems = treeChildrenSelected.childNodes;
            var len = treeItems.length;

            for ( var i = 0; i < len; i++ ) {
                if ( treeItems[i].firstChild.firstChild.getAttribute("label") ===
                        this.selectedLocation.location ) {
                    alert(strbundle.getString("alert.warning.duplicate.location"));
                    return;
                }
            }
        }

        this.selectedLocation.currentlyDst = false;
        var offset = this.zoneOffset[index];
        this.selectedLocation.offsetMilli = this.convertSignedClockToMilli(offset);
        this.selectedLocation.localTimeZone = this.zoneFormat[index];

        // Get dst nodes, attributes for selected zone.
        if ( this.zoneRule[index] != "N/A" ) {
            this.selectedLocation.observesDst = true;
            this.selectedLocation.rule = this.zoneRule[index];
            var countryIndex = this.selectedCountryIndex + 1;
            var zoneIndex = index + 1;

            var exp = "/countrytimezones/country[" + countryIndex + "]/timezone[" + zoneIndex + "]/*";

            var dstNodeSet = this.countriesTimeZonesDoc.
                    evaluate(exp, this.countriesTimeZonesDoc, null, XPathResult.ANY_TYPE, null);
            var dstNode = dstNodeSet.iterateNext();

            if ( !dstNode ) {
                alert(strbundle.getString("alert.error.missing.dstrules") + " " + this.zoneRule[index]);
                return;
            }

            while ( dstNode ) {
                this.getDstAttrs(dstNode, dstNode.nodeName);
                dstNode = dstNodeSet.iterateNext();
            }

            this.selectedLocation.dstOffsetMilli = this.selectedLocation.offsetMilli + this.convertSignedClockToMilli(this.saveStart);
            this.selectedLocation.currentlyDst = this.currentlyDst(this.selectedLocation, 0);
            this.formatLocalZone(this.selectedLocation, this.letterStart, this.letterEnd);
            this.buildDstDisplay();
        }
        else {
            this.selectedLocation.rule = "";
            this.selectedLocation.observesDst = false;
            this.selectedLocation.dstOffsetMilli = 0;
            this.selectedLocation.dstLocalTimeZone = "";
            this.selectedLocation.dstStartMilli = [0,0,0];
            this.selectedLocation.dstEndMilli = [0,0,0];
            this.selectedDisplay.dstStart = strbundle.getString("msg.not.applicable");
            this.selectedDisplay.dstEnd = strbundle.getString("msg.not.applicable");
            this.selectedDisplay.localTimeZone = this.selectedLocation.localTimeZone;
        }

        this.selectedDisplay.offset = this.selectedLocation.currentlyDst ?
                this.convertMilliToClock(this.selectedLocation.dstOffsetMilli) + "**" :
                offset;

        this.pushCellData(treeChildrenSelected);

        // Select the added row.
        var rowIndex = treeItems ? treeItems.length - 1 : 0;
        treeSelected.view.selection.select(rowIndex);

        this.ensureRowVisible(treeChildrenSelected);
    },

    // Add cell data to array.

    pushCellData: function(treeChildrenSelected) {
        var addData = [];

        addData.push(this.selectedDisplay.location);
        addData.push(this.selectedDisplay.offset);
        addData.push(this.selectedDisplay.dstStart);
        addData.push(this.selectedDisplay.dstEnd);
        addData.push(this.selectedDisplay.localTimeZone);

        this.addTreeRow(treeChildrenSelected, addData);
     },

    // Add row to a tree.

    addTreeRow: function(treeChildren, addData) {
        // Convert selectedLocation object to JSON string.
        var json = Components.classes["@mozilla.org/dom/json;1"].
                createInstance(Components.interfaces.nsIJSON);
        var selectedLocationString = json.encode(this.selectedLocation);

        var item = document.createElement("treeitem");
        var row = document.createElement("treerow");
        var cell;

        for ( var i in addData ) {
            cell = document.createElement("treecell");
            cell.setAttribute("label", addData[i]);

            if ( i === "0" ) {
                // Stash the JSON string to be saved in the pref.
                cell.setAttribute("value", selectedLocationString);
            }
            else {
                cell.setAttribute("editable", "false");
            }

            row.appendChild(cell);
        }

       // Add the treerow onto treeitem, then treeitem onto treechildren.
        item.appendChild(row);
        treeChildren.appendChild(item);
    },

    // Ensure the just-added row is visible.

    ensureRowVisible: function(treeChildrenSelected) {
        var treeItemsSelected = treeChildrenSelected.childNodes;

        if ( treeItemsSelected.length > 4 ) {
            var boxobject = document.getElementById("simclk-treeSelected").boxObject;

            boxobject.QueryInterface(Components.interfaces.nsITreeBoxObject);
            boxobject.ensureRowIsVisible(treeItemsSelected.length - 1);
        }
    },

    // Get dst node attributes.
    // Return values used by update routine.

    getDstAttrs: function(dstNode, nodeName) {
        var save = dstNode.getAttribute("save");
        var letter = dstNode.getAttribute("letter");
        var dstMilli = dstNode.getAttribute("dates").split(",");

        if ( nodeName == "dstend" ) {
            this.letterEnd = letter;
            this.saveEnd = save;

            for ( var i in dstMilli ) {
                this.selectedLocation.dstEndMilli[i] = parseInt(dstMilli[i], 10);
            }
        }
        else {
            this.letterStart = letter;
            this.saveStart = save;

            for ( var j in dstMilli ) {
                this.selectedLocation.dstStartMilli[j] = parseInt(dstMilli[j], 10);
            }


        }
    },

    // Check if dst in effect for given location.
    // Checks for given milliTime, or for current time.

    currentlyDst: function(location, milliTime) {
        var currTimeMilli = (milliTime > 0) ? milliTime : new Date().getTime();

        for ( var i in location.dstStartMilli ) {
            if ( location.dstStartMilli[i] === 0 ) {
                continue;
            }

            if ( currTimeMilli >= location.dstStartMilli[i] &&
                 (currTimeMilli < location.dstEndMilli[i] || location.dstEndMilli[i] === 0) ) {
                return true;
            }
        }

        return false;
    },

    // Construct dst start and end dates for display in Selected tree.

    buildDstDisplay: function() {
        var currTimeMilli = new Date().getTime();

        // The XML file contains 3 sets of dst start/end dates for 3 consecutive years.
        // Display the set in effect or if not in effect, the upcoming set.
        for ( var i in this.selectedLocation.dstEndMilli ) {
            if ( currTimeMilli < this.selectedLocation.dstEndMilli[i] ) {
                this.selectedDisplay.dstStart = this.formatDate(this.selectedLocation.dstStartMilli[i], "start");
                this.selectedDisplay.dstEnd = this.formatDate(this.selectedLocation.dstEndMilli[i], "end");
                return;
            }
        }

        for ( var j = 2; j >= 0; j-- ) {
            if ( this.selectedLocation.dstStartMilli[j] > 0 ||
                 j === 0 ) {
                this.selectedDisplay.dstStart = this.formatDate(this.selectedLocation.dstStartMilli[j], "start");
                this.selectedDisplay.dstEnd = this.formatDate(this.selectedLocation.dstEndMilli[j], "end");
                return;
            }
        }
    },

    // Formats a millisec date for display.

    formatDate: function(milliDate, startEnd) {
        var strbundle = document.getElementById("simclk-strings");
        var format;

        if ( milliDate === 0) {
            return strbundle.getString("msg.not.applicable");
        }

        if ( startEnd === "start" ) {
            format = this.selectedLocation.localTimeZone;
            milliDate += this.selectedLocation.offsetMilli;
        }
        else {
            format = this.selectedLocation.dstLocalTimeZone;
            milliDate += this.selectedLocation.dstOffsetMilli;
        }

        var date = new Date(milliDate);
        var dow = this.getDow(date.getUTCDay());
        var month = this.getMonth(date.getUTCMonth());
        var day = date.getUTCDate();
        var year = date.getUTCFullYear();
        var hours = date.getUTCHours();
        var minutes = this.pad(date.getUTCMinutes());

        // Special case of dst time 24:00, we use 23:59:59.
        var seconds = (date.getUTCSeconds() === 59) ? ":59" : "";

        return (dow + ", " + month + " " + day + " " + year +
                " " + hours + ":" + minutes + seconds + " " + format);
    },

    // Convert numeric value to day of week.

    getDow: function(dow) {
        var strbundle = document.getElementById("simclk-strings");
        var daysOfWeek = [
                strbundle.getString("display.dow.short.sunday"),
                strbundle.getString("display.dow.short.monday"),
                strbundle.getString("display.dow.short.tuesday"),
                strbundle.getString("display.dow.short.wednesday"),
                strbundle.getString("display.dow.short.thursday"),
                strbundle.getString("display.dow.short.friday"),
                strbundle.getString("display.dow.short.saturday"),
                strbundle.getString("display.dow.daily"),
                strbundle.getString("display.dow.short.weekdays"),
                strbundle.getString("display.dow.short.weekend")
                ];

        if ( daysOfWeek[dow] != null ) {
            // When undefined (ie. daysOfWeek[-1]) is compared to null, the result is true.
            return daysOfWeek[dow];
        }
        else if ( dow === -1 ) {
            return "";
        }
        else {
            alert(strbundle.getString("alert.error.invalid.dow"));
            return "Err";
        }
    },

    // Convert numeric month to string.

    getMonth: function(mo) {
        var strbundle = document.getElementById("simclk-strings");
        var months = [
                strbundle.getString("display.month.short.january"),
                strbundle.getString("display.month.short.february"),
                strbundle.getString("display.month.short.march"),
                strbundle.getString("display.month.short.april"),
                strbundle.getString("display.month.short.may"),
                strbundle.getString("display.month.short.june"),
                strbundle.getString("display.month.short.july"),
                strbundle.getString("display.month.short.august"),
                strbundle.getString("display.month.short.september"),
                strbundle.getString("display.month.short.october"),
                strbundle.getString("display.month.short.november"),
                strbundle.getString("display.month.short.december")
                ];

        if ( months[mo] != null ) {
            // When undefined is compared to null, the result is true.
            return months[mo];
        }
        else {
            alert (strbundle.getString("alert.error.invalid.month"));
            return "Err";
        }
    },

    // For locations that observe dst, the local time zone usually (not always)
    // contains a placeholder to toggle the field between standard and
    // dst formats (eg. EST, EDT). Replace placeholder with appropriate letter.
    // If the difference is more complex, they are seperated by a slash instead
    // of using a placeholder.
    // Note non-primitive objects get passed by reference.

    formatLocalZone: function(location, letterStart, letterEnd) {
        if ( location.localTimeZone.match("%s") ) {
            if ( letterStart != "-") {
                location.dstLocalTimeZone =
                        location.localTimeZone.replace(/%s/, letterStart);
            }
            else {
                location.dstLocalTimeZone =
                        location.localTimeZone.replace(/%s/, "");
            }

            if ( letterEnd != "-") {
                location.localTimeZone =
                        location.localTimeZone.replace(/%s/, letterEnd);
            }
            else {
                location.localTimeZone =
                        location.localTimeZone.replace(/%s/, "");
            }

            this.selectedDisplay.localTimeZone = location.currentlyDst ?
                    location.dstLocalTimeZone : location.localTimeZone;
        }
        else if ( location.localTimeZone.match("/") ) {
            var formats = location.localTimeZone.split("/");
            location.localTimeZone = formats[0];
            location.dstLocalTimeZone = formats[1];

            this.selectedDisplay.localTimeZone = location.currentlyDst ?
                    location.dstLocalTimeZone : location.localTimeZone;
        }
        else {
            this.selectedDisplay.localTimeZone = location.localTimeZone;
            location.dstLocalTimeZone = location.localTimeZone;
        }
    },

    // Convert millisecs to signed clock.

    convertMilliToClock: function(milli) {
        var hrs;
        var mins = Math.floor(Math.abs(milli) / ( 1000 * 60 )) % 60;

        if ( milli >= 0 ) {
            hrs = Math.floor(milli / ( 1000 * 60 * 60 ));
        }
        else {
            hrs = Math.ceil(milli / ( 1000 * 60 * 60 ));
        }

         if ( hrs >= 0 ) {
            return ( "+" + hrs + ":" + this.pad(mins) );
        }
        else
        {
            return ( hrs + ":" + this.pad((mins)) );
        }
    },

    // Convert a signed clock to millisecs.

    convertSignedClockToMilli: function(clock) {
        // Just in case -0:30, when parseInt(clock) returns 0.
        while ( clock.substring(0,1) == " ") {
            clock = clock.substring(1);
        }

        if ( clock.charAt(0) != "-" ) {
            return ( ( parseInt(clock, 10) * 1000 * 60 * 60 ) +
                     ( parseInt(clock.substr(clock.indexOf(":") + 1), 10) * 60 * 1000 ) );
        }
        else {
            return ( ( parseInt(clock, 10) * 1000 * 60 * 60 ) -
                     ( parseInt(clock.substr(clock.indexOf(":") + 1), 10) * 60 * 1000 ) );
        }
    },

    // Pad a number with zero.

    pad: function(num) {
        return (num < 10 ? "0" + num : num);
    },

    // Called to delete selection(s) from tree.

    onDelete: function(aTree) {
        var strbundle = document.getElementById("simclk-strings");
        var tree = document.getElementById(aTree);

        if ( tree.currentIndex < 0 ) {
            // User didn't make a selection.
            alert(strbundle.getString("alert.warning.no.item.selected"));
        }
        else {
            var start = {};
            var end = {};
            var treeItem, location;
            var numRanges = tree.view.selection.getRangeCount();

            // Start with the last selection range and work up.
            for ( var i = numRanges - 1; i >= 0; i-- ) {
                tree.view.selection.getRangeAt(i, start, end);

                // Within a range, delete from the bottom up.
                for ( var j = end.value; j >= start.value; j-- ) {
                    location = tree.view.
                            getCellText(j, tree.columns.getNamedColumn("simclk-colCity"));

                    if ( this.checkLocationForNots(location) ) {
                        treeItem = tree.view.getItemAtIndex(j);
                        treeItem.parentNode.removeChild(treeItem);
                    }
                }
            }
        }
    },

    // Called when user clicks Del All. Also Reset.

    onDeleteAll: function(aTreeChildren, mode) {
        var treeChildren = document.getElementById(aTreeChildren);
        var location;

        // Remove from the bottom.
        while ( treeChildren.hasChildNodes() ) {
            location = treeChildren.lastChild.firstChild.firstChild.getAttribute("label");

            // Don't check for nots. when doing a Reset.
            if ( mode != "delAll" ||
                 this.checkLocationForNots(location) ) {
                treeChildren.removeChild(treeChildren.lastChild);
            }
            else {
                return;
            }
        }
    },

    // If the location to be deleted has notifications, inform user.

    checkLocationForNots: function(location) {
        var strbundle = document.getElementById("simclk-strings");
        var button = true;

        for ( var i in this.notifyTable ) {

            if ( !this.notifyTable[i].completed &&
                 this.notifyTable[i].location == location ) {

                button = confirm(location + " " +
                        strbundle.getString("confirm.cancel.notifications.line1") +
                        "\n" +
                        strbundle.getString("confirm.cancel.notifications.line2"));

                if ( button ) {
                    // User clicked OK.
                    this.deleteNotsForLocation.push(location);
                }

                break;
            }
        }

        return button;
    },

    // Called when user clicks Enter.

    onEnter: function() {
        var strbundle = document.getElementById("simclk-strings");
        var city = document.getElementById("simclk-tboxCity").value;

        if ( !city ) {
            alert(strbundle.getString("alert.warning.no.city.entered"));
            return;
        }

        var lboxCountries = document.getElementById("simclk-lboxCountries");
        var country;

        if ( lboxCountries.selectedIndex < 0 ) {
            country = "";
        }
        else {
            country = lboxCountries.getItemAtIndex(this.selectedCountryIndex).label;
        }

        this.openLink("http://worldtimeengine.com/doTime/" + city + " " + country);
    },

    // Called when user clicks OK. Write locations preference,
    // then call build table routine.

    onOk: function() {
        var json = Components.classes["@mozilla.org/dom/json;1"].
                createInstance(Components.interfaces.nsIJSON);

        var Application = Components.classes["@mozilla.org/fuel/application;1"].
                getService(Components.interfaces.fuelIApplication);

        for ( var i in this.deleteNotsForLocation ) {
            for ( var j in this.notifyTable ) {
                if ( !this.notifyTable[j].completed &&
                     this.notifyTable[j].location == this.deleteNotsForLocation[i] ) {
                    this.notifyTable[j].completed = true;
                }
            }
        }

        if ( this.deleteNotsForLocation ) {
            Application.storage.set("notifyTable", this.notifyTable);
            SimpleClocksNotify.updateNotifyPref();
        }

        var treeChildrenSelected = document.getElementById("simclk-treeSelectedItems");
        var prefString = "";
        var row, jsonString, selectedLocation;

        if ( treeChildrenSelected.hasChildNodes() ) {
            var treeItems = treeChildrenSelected.childNodes;
            var len = treeItems.length;

            for ( var k = 0; k < len; k++ ) {
                // In case user edited location.
                row = treeItems[k].firstChild;

                // Get the full location object that we stashed.
                jsonString = row.firstChild.getAttribute("value");
                selectedLocation = json.decode(jsonString);
                selectedLocation.location = row.firstChild.getAttribute("label");
                prefString += json.encode(selectedLocation) + this.separator;
            }
        }

        // Remove last separator.
        if ( prefString ) {
            prefString = prefString.substring(0, prefString.length - 1);
        }

        this.setLocationsPref(prefString);
        this.buildLocationsTable();
    },

    // Called on startup or when user clicks OK.
    // Build locations table from preference.

    buildLocationsTable: function() {
        var Application = Components.classes["@mozilla.org/fuel/application;1"].
                getService(Components.interfaces.fuelIApplication);

        var json = Components.classes["@mozilla.org/dom/json;1"].
                createInstance(Components.interfaces.nsIJSON);

        var locations = this.getLocationsPref();

        if ( locations ) {
            for ( var i in locations ) {
                this.selectedLocation = json.decode(locations[i]);
                this.selectedLocationsTable[i] = this.selectedLocation;
            }
        }
        else {
            this.selectedLocationsTable.length = 0;
        }

        Application.storage.set("locationsTable", this.selectedLocationsTable);
    },

    // Returns an array of location JSON strings
    // from the locations preference.

    getLocationsPref: function() {
        var prefs = Components.classes["@mozilla.org/preferences-service;1"].
                getService(Components.interfaces.nsIPrefService).
                getBranch("extensions.simpletimerClocks@grbradt.org.");

        var locationsPref = prefs.getComplexValue("locations",
                Components.interfaces.nsISupportsString).data;

        if ( locationsPref ) {
            return locationsPref.split("\x1D"); // The not. separator we use.
        }
        else {
            return "";
        }
    },

    // Set the locations preference.

    setLocationsPref: function(prefString) {
        var prefs = Components.classes["@mozilla.org/preferences-service;1"]
                .getService(Components.interfaces.nsIPrefService)
                .getBranch("extensions.simpletimerClocks@grbradt.org.");
        var str = Components.classes["@mozilla.org/supports-string;1"].
                createInstance(Components.interfaces.nsISupportsString);

        str.data = prefString;
        prefs.setComplexValue("locations", Components.interfaces.nsISupportsString, str);
    },

    // Called from SimpleClocks.
    // If the passed location is being tracked, return the location object.

    searchLocationsTable: function(location) {
        var Application = Components.classes["@mozilla.org/fuel/application;1"].
                getService(Components.interfaces.fuelIApplication);

        // An array of selectedLocation objects (see SimpleClocksLocations).
        var locationsTable = Application.storage.get("locationsTable", "ERR");

        if ( locationsTable === "ERR" ) {
            alert(strbundle.getString("alert.error.loading.locations"));
            return null;
        }

        for ( var i in locationsTable ) {
            if ( locationsTable[i].location == location ) {
                return locationsTable[i];
            }
        }

        return null;
    },

    // Called from SimpleClocks when building tooltip.

    getLocationTime: function(location, clock24Hr) {
        // First convert curr date to locale date by adding in locale UTC offset.
        var currDate = new Date();
        var currDateMilli = currDate.getTime();
        var currOffsetMilli = currDate.getTimezoneOffset() * 60 * 1000;
        var localeDateMilli = currDateMilli + location.offsetMilli;
        var dstInEffect = false;

        if ( location.observesDst ) {
            if ( SimpleClocksLocations.currentlyDst(location, 0) ) {
                dstInEffect = true;
                localeDateMilli = currDateMilli + location.dstOffsetMilli;
            }
        }

        var localeDate = new Date(localeDateMilli);

        // Start fiddling to determine if user is 12 or 24 clock format,
        // and get user's AM/PM chars for 12 hr clock.
        var amTimeStr = new Date("December 23, 2007 01:00:00").toLocaleTimeString().split(" ");
        var pmTimeStr = new Date("December 23, 2007 13:00:00").toLocaleTimeString().split(" ");

        // Used for statusbar tooltip.
        var returnTime;

        // Used for notifications tooltip.
        var timeDiffHrs = ( localeDateMilli - currDateMilli + currOffsetMilli ) / 1000 / 60 / 60;

        if ( clock24Hr || amTimeStr.length === 1 ) {
            // 24 hr clock.
            if ( dstInEffect ) {
                returnTime = (this.pad(localeDate.getUTCHours()) + ":" + this.pad(localeDate.getUTCMinutes()) +
                        "**" + ", " + this.getDow(localeDate.getUTCDay()));
            }
            else {
                returnTime = (this.pad(localeDate.getUTCHours()) + ":" + this.pad(localeDate.getUTCMinutes()) +
                        ", " + this.getDow(localeDate.getUTCDay()));
            }
        }
        else {
            // 12 hr clock.
            var hrs = localeDate.getUTCHours();
            hrs = ( hrs > 12 ) ? hrs - 12 : hrs;
            hrs = ( hrs === 0 ) ? 12 : hrs;
            var ampm;

            if ( amTimeStr[0].indexOf(":") >= 0 ) {
                // Time value precedes AM/PM chars (typical).
                // timeStr[1] will be AM/PM chars.
                ampm = ( localeDate.getUTCHours() < 12 ) ? amTimeStr[1] : pmTimeStr[1];

                if ( dstInEffect ) {
                    returnTime = (hrs + ":" + this.pad(localeDate.getUTCMinutes()) +
                            " " + ampm + "**" + ", " + this.getDow(localeDate.getUTCDay()));
                }
                else {
                    returnTime = (hrs + ":" + this.pad(localeDate.getUTCMinutes()) +
                            " " + ampm + ", " + this.getDow(localeDate.getUTCDay()));
                }
            }
            else {
                // AM/PM chars precede time value eg. Korean.
                // timeStr[0] will be AM/PM chars.
                ampm = ( localeDate.getUTCHours() < 12 ) ? amTimeStr[0] : pmTimeStr[0];

                if ( dstInEffect ) {
                    returnTime = (ampm + " " + this.pad(hrs) + ":" + this.pad(localeDate.getUTCMinutes()) +
                            "**" + ", " + this.getDow(localeDate.getUTCDay()));
                }
                else {
                    returnTime = (ampm + " " + this.pad(hrs) + ":" + this.pad(localeDate.getUTCMinutes()) +
                            ", " + this.getDow(localeDate.getUTCDay()));
                }
            }
        }

        return {time: returnTime, timeDiffHrs: timeDiffHrs};
    },

    // Called during startup when tz data file has been updated.

    getXmlDoc: function() {
        this.readCountriesTimeZonesFile("update");
    },

    // Called during startup when tz data file has been updated.
    // Document retrieved, now update locations pref.

    updateLocationsPref: function() {
        var strbundle = document.getElementById("simclk-strings");
        var prefString = "";

        var Application = Components.classes["@mozilla.org/fuel/application;1"].
                getService(Components.interfaces.fuelIApplication);
        var json = Components.classes["@mozilla.org/dom/json;1"].
                createInstance(Components.interfaces.nsIJSON);

        var zoneOffset, exp, dstNodeSet, dstNode, save, saveStart, letterStart, letterEnd;

        var countriesDoc = Application.storage.get("xmlCountriesTimeZonesDoc", "No doc");

        // Get the pref; locations is an array of location objects.
        var locations = SimpleClocksLocations.getLocationsPref();

        for ( var i in locations ) {
            this.selectedLocation = json.decode(locations[i]);

            exp = "/countrytimezones/country[@code='" +
                    this.selectedLocation.countryCode + "']/timezone[@name='" +
                    this.selectedLocation.timezone + "']/@offset";
            zoneOffset = SimpleClocksLocations.getXMLSingleNodeValue(exp, countriesDoc);

            exp = "/countrytimezones/country[@code='" +
                    this.selectedLocation.countryCode + "']/timezone[@name='" +
                    this.selectedLocation.timezone + "']/@format";
            this.selectedLocation.localTimeZone = SimpleClocksLocations.getXMLSingleNodeValue(exp, countriesDoc);

            exp = "/countrytimezones/country[@code='" +
                    this.selectedLocation.countryCode + "']/timezone[@name='" +
                    this.selectedLocation.timezone + "']/@rule";
            this.selectedLocation.rule = SimpleClocksLocations.getXMLSingleNodeValue(exp, countriesDoc);

            if ( !zoneOffset ||
                 !this.selectedLocation.localTimeZone ||
                 this.selectedLocation.rule === null ) {
                alert(strbundle.getString("alert.error.missing.node1") + " " +
                      this.selectedLocation.countryCode + " " +
                      strbundle.getString("alert.error.missing.node2") + " " +
                      this.selectedLocation.timezone);

                return;
            }

            this.selectedLocation.offsetMilli = this.convertSignedClockToMilli(zoneOffset);

            // Get dst nodes, attributes for selected zone.
            if ( this.selectedLocation.rule !== "") {
                this.selectedLocation.observesDst = true;

                exp = "/countrytimezones/country[@code='" +
                        this.selectedLocation.countryCode + "']/timezone[@name='" +
                        this.selectedLocation.timezone + "']/*";
                dstNodeSet = countriesDoc.
                        evaluate(exp, countriesDoc, null, XPathResult.ANY_TYPE, null);
                dstNode = dstNodeSet.iterateNext();

                if ( !dstNode ) {
                    alert(strbundle.getString("alert.error.missing.dstrules") + " " +
                          this.selectedLocation.rule);
                    return;
                }

                while ( dstNode )
                {
                    SimpleClocksLocations.getDstAttrs(dstNode, dstNode.nodeName);
                    dstNode = dstNodeSet.iterateNext();
                }

                SimpleClocksLocations.formatLocalZone(this.selectedLocation, this.letterStart, this.letterEnd);
                this.selectedLocation.dstOffsetMilli =
                        this.selectedLocation.offsetMilli + SimpleClocksLocations.convertSignedClockToMilli(this.saveStart);
            }
            else {
                this.selectedLocation.observesDst = false;
                this.selectedLocation.dstOffsetMilli = 0;
                this.selectedLocation.dstLocalTimeZone = "";
                this.selectedLocation.dstStartMilli = [0,0,0];
                this.selectedLocation.dstEndMilli = [0,0,0];
            }

            prefString += json.encode(this.selectedLocation) + "\x1D";
        }

        // Remove last separator.
        prefString = prefString.substring(0, prefString.length - 1);

        // And update the locations pref.
        SimpleClocksLocations.setLocationsPref(prefString);
    },

    // Open page in new tab, minimize dialog.
    // This function developed by Devon Jensen, from Download Statusbar.

    openLink: function(aPage) {
        var wm = Components.classes["@mozilla.org/appshell/window-mediator;1"].getService();
        var wmed = wm.QueryInterface(Components.interfaces.nsIWindowMediator);
        var win = wmed.getMostRecentWindow("navigator:browser");

        if ( !win ) {
            win = window.openDialog("chrome://browser/content/browser.xul",
                                    "_blank",
                                    "chrome,all,dialog=no",
                                    aPage, null, null);
        }
        else {
            var content = win.document.getElementById("content");
            content.selectedTab = content.addTab(aPage);
            window.minimize();
        }
    },

    // Dump to console.

    dumpLocationObject: function(location) {
        this.debug("\nLocation data:");
        this.debug("Country code: " + location.countryCode);
        this.debug("tz time zone: " + location.timezone);
        this.debug("Location: " + location.location);
        this.debug("Rule: " + location.rule);
        this.debug("Observes dst?: " + location.observesDst);
        this.debug("Currently dst?: " + location.currentlyDst);
        this.debug("UTC offset in millisecs: " + location.offsetMilli);
        this.debug("UTC offset in millisecs for dst: " + location.dstOffsetMilli);
        this.debug("Local time zone: " + location.localTimeZone);
        this.debug("Local time zone for dst: " + location.dstLocalTimeZone);
        this.debug("dst millisec start time: " + location.dstStartMilli);
        this.debug("dst millisec end time: " + location.dstEndMilli);
    },

    // Debug messages to console.

    debug: function(aMsg) {
        setTimeout(function() { throw new Error("[debug] " + aMsg); }, 0);
    }
};